package com.anecdote.ideaplugins.changesbar;

/**
 * Copyright 2006 Nathan Brown
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.intellij.ide.DataManager;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.popup.JBPopupFactory;
import com.intellij.openapi.ui.popup.ListPopup;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.vcs.changes.*;
import com.intellij.openapi.wm.StatusBar;
import com.intellij.openapi.wm.ToolWindowManager;
import com.intellij.ui.awt.RelativePoint;
import com.intellij.util.ui.UIUtil;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.StringWriter;
import java.util.Collection;
import java.util.List;

class ChangesBar extends JPanel implements ChangeListListener
{
  public static final String VERSION = "1.4";
  private static final Icon ARROW_ICON = IconLoader.getIcon("/general/comboArrow.png");
  private final ChangeListSelectionButton _defaultListSelectionButton = new ChangeListSelectionButton();
  private ChangeListManager _changeListManager;
  private ChangeList _defaultChangeList;
  private static final Icon CHANGES_ICON = IconLoader.getIcon("/general/toolWindowChanges.png");
  static final Icon WARNING_ICON = IconLoader.getIcon("/com/anecdote/ideaplugins/changesbar/warning.gif");
  private final AnAction _newChangelistAction = new NewChangeListAction(this);
  private final JButton _warningButton = new JButton(WARNING_ICON);
  private final Project _project;
  private final StatusBar _statusBar;
  private boolean _showWarning;
  private Runnable _queuedWarning;
  private ConflictPopup.InternalPopupPane _popup;
  private static final Color POPUP_COLOUR = new Color(223, 172, 82);
  private static final JLabel POPUP_CONTENT = new JLabel("<html>Changelist Collision Detected!<BR>" +
                                                         "Click here to view...</html>", WARNING_ICON,
                                                                                         SwingConstants.LEFT);

  ChangesBar(final Project project, StatusBar statusBar)
  {
    super(new BorderLayout());
    _project = project;
    _statusBar = statusBar;
    add(_warningButton, BorderLayout.WEST);
    _warningButton.setVisible(false);
    _warningButton.setFocusable(false);
    _warningButton.setMargin(new Insets(1, 1, 0, 1));
    _warningButton.putClientProperty("Quaqua.Button.style", "placard");
    _warningButton.setToolTipText("A Changelist Collision has been detected.\nClick to Clear Collision Warning");
    _warningButton.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        showChangesToolwindow(project);
      }
    });
    POPUP_CONTENT.addMouseListener(new MouseAdapter()
    {
      @Override
      public void mouseClicked(MouseEvent e)
      {
        showChangesToolwindow(project);
      }
    });
    add(_defaultListSelectionButton, BorderLayout.CENTER);
    updateButtonText(_defaultChangeList);
  }

  private void showChangesToolwindow(Project project)
  {
    ToolWindowManager.getInstance(project).getToolWindow(ChangesBarProjectComponent.CHANGES_TOOLWINDOW_NAME).show(null);
    setWarningVisible(false);
    SwingUtilities.getWindowAncestor(this).validate();
    SwingUtilities.getWindowAncestor(this).invalidate();
    SwingUtilities.getWindowAncestor(this).repaint();
  }

  void setChangeListManager(ChangeListManager changeListManager)
  {
    if (_changeListManager == changeListManager) {
      return;
    }
    if (_changeListManager != null) {
      _changeListManager.removeChangeListListener(this);
    }
    _changeListManager = changeListManager;
    defaultListChanged(_changeListManager != null ? _changeListManager.getDefaultChangeList() : null);
    if (_changeListManager != null) {
      _changeListManager.addChangeListListener(this);
    }
  }

  public void changeListAdded(ChangeList list)
  {
  }

  public void changeListRemoved(ChangeList list)
  {
    if (_defaultChangeList.equals(list)) {
      defaultListChanged(null);
    }
  }

  public void changeListChanged(ChangeList list)
  {
  }

  public void changeListRenamed(ChangeList list, String oldName)
  {
    if (list.equals(_defaultChangeList)) {
      updateButtonText(list);
    }
  }

  public void changeListCommentChanged(ChangeList list, String oldComment)
  {
  }

  private void updateButtonText(ChangeList list)
  {
    if (list != null) {
      _defaultListSelectionButton.setText(list.getName() + " (" + list.getChanges().size() + ')');
    } else {
      _defaultListSelectionButton.setText("[No List]");
    }
  }

  public void changesMoved(Collection<Change> changes, ChangeList fromList, ChangeList toList)
  {
    updateButtonText(_defaultChangeList);
  }

  public void defaultListChanged(ChangeList newDefaultList)
  {
    _defaultChangeList = newDefaultList;
    updateButtonText(newDefaultList);
  }

  public void unchangedFileStatusChanged()
  {
    updateButtonText(_defaultChangeList);
  }

  public void changeListUpdateDone()
  {
    updateButtonText(_defaultChangeList);
  }

  private DefaultActionGroup createSelectionActionGroup()
  {
    final DefaultActionGroup actionGroup = new DefaultActionGroup();
    if (_changeListManager != null) {
      final List<LocalChangeList> changeLists = _changeListManager.getChangeLists();
      for (final LocalChangeList changeList : changeLists) {
        if (!ChangesBarProjectComponent.isConflictChangeList(changeList)) {
          actionGroup.add(new ChangeListSelectionAction(changeList));
        }
      }
    }
    actionGroup.add(_newChangelistAction);
    return actionGroup;
  }

  void setWarningVisible(boolean visible)
  {
    _showWarning = visible;
    if (visible) {
      if (_queuedWarning == null) {
        _queuedWarning = new Runnable()
        {
          public void run()
          {
            if (_showWarning) {
              _warningButton.setVisible(true);
              getParent().getParent().getParent().validate();
              if (ChangesBarApplicationComponent.isShowPopup() && _popup == null) {
                _popup = ConflictPopup.popup(_warningButton, "Changelist Collision Detected!", WARNING_ICON,
                                             "Click here to clear this alert and open the Changes toolwindow.",
                                             new AbstractAction()
                                             {
                                               public void actionPerformed(ActionEvent e)
                                               {
                                                 showChangesToolwindow(_project);
                                               }
                                             }, 0);
                _popup.addPropertyChangeListener(ConflictPopup.InternalPopupPane.VISIBLE, new PropertyChangeListener()
                {
                  public void propertyChange(PropertyChangeEvent evt)
                  {
                    if (Boolean.FALSE.equals(evt.getNewValue())) {
                      _popup.removePropertyChangeListener(ConflictPopup.InternalPopupPane.VISIBLE, this);
                      _popup = null;
                    }
                  }
                });
//                _statusBar.fireNotificationPopup(POPUP_CONTENT, POPUP_COLOUR);
              }
            }
            _queuedWarning = null;
          }
        };
      }
      SwingUtilities.invokeLater(_queuedWarning);
    } else {
      _warningButton.setVisible(false);
      if (_popup != null) {
        _popup.setVisible(false);
      }
    }
  }

  Project getProject()
  {
    return _project;
  }

  private class ChangeListSelectionButton extends JButton
  {
    private boolean _myForcePressed = false;
    private static final int MAX_ROW_COUNT = 30;
    private static final int PREF_BUTTON_HEIGHT = 21;

    private ChangeListSelectionButton()
    {
      super(CHANGES_ICON);
      setModel(new MyButtonModel());
      setHorizontalAlignment(SwingConstants.LEFT);
      setFocusable(false);
      final Insets margins = getMargin();
      setMargin(new Insets(margins.top, 2, margins.bottom, 2));
      addActionListener(
        new ActionListener()
        {
          public void actionPerformed(ActionEvent e)
          {
            if (!_myForcePressed) {
              showPopup();
            }
//            ToolWindow window = ToolWindowManager.getInstance(_project)
//              .getToolWindow(ChangesBarProjectComponent.CHANGES_TOOLWINDOW_NAME);
//            if (window != null) {
//              JComponent changesToolwindowComponent = window.getComponent();
//              dumpComponentHierarchy(0, changesToolwindowComponent);
//            }
          }
        }
      );

      //noinspection HardCodedStringLiteral
      putClientProperty("Quaqua.Button.style", "placard");
    }

    private void showPopup()
    {
      final DefaultActionGroup group = createSelectionActionGroup();
      _myForcePressed = true;
      repaint();

      final Runnable onDispose = new Runnable()
      {
        public void run()
        {
          // give button chance to handle action listener
          SwingUtilities.invokeLater(new Runnable()
          {
            public void run()
            {
              _myForcePressed = false;
            }
          });
          repaint();
        }
      };
      final ListPopup popup = JBPopupFactory.getInstance().createActionGroupPopup(
        "Changelists", group,
        DataManager.getInstance().getDataContext(),
        JBPopupFactory.ActionSelectionAid.SPEEDSEARCH,
        false,
        onDispose,
        MAX_ROW_COUNT);
      final Dimension preferredSize = popup.getContent().getPreferredSize();
      preferredSize.width = Math.max(preferredSize.width, getSize().width);
      popup.getContent().setPreferredSize(preferredSize);
      final Point point = getBounds().getLocation();
      point.y -= preferredSize.height + getHeight();
      popup.show(new RelativePoint(this, point));
    }

    @Override
    public void setText(String text)
    {
      super.setText(text);
      updateButtonSize();
    }

    @Override
    public void updateUI()
    {
      super.updateUI();
      if (UIUtil.isMotifLookAndFeel()) {
        setBorder(BorderFactory.createEtchedBorder());
      } else {
        setBorder(UIUtil.getButtonBorder());
      }
    }

    private class MyButtonModel extends DefaultButtonModel
    {
      @Override
      public boolean isPressed()
      {
        return _myForcePressed || super.isPressed();
      }

      @Override
      public boolean isArmed()
      {
        return _myForcePressed || super.isArmed();
      }
    }

    @Override
    public final void paint(Graphics g)
    {
      super.paint(g);
      final Dimension size = getSize();
      final String text = getText();
      final boolean isEmpty = getIcon() == null && (text == null || text.trim().length() == 0);
      final int x = isEmpty ?
                    (size.width - ARROW_ICON.getIconWidth()) >> 1 :
                    size.width - ARROW_ICON.getIconWidth() - 2;
      ARROW_ICON.paintIcon(null, g, x, (size.height - ARROW_ICON.getIconHeight()) >> 1);
    }

    private void updateButtonSize()
    {
      int width;
      final String text = getText();
      if ((text == null || text.trim().length() == 0) && getIcon() == null) {
        width = ARROW_ICON.getIconWidth() + 10;
      } else {
        width = getUI().getPreferredSize(this).width + ARROW_ICON.getIconWidth() + 2;
      }
      setPreferredSize(new Dimension(width, PREF_BUTTON_HEIGHT));
    }
  }

  private class ChangeListSelectionAction extends ToggleAction
  {
    private final LocalChangeList _changeList;

    private ChangeListSelectionAction(LocalChangeList changeList)
    {
      super(changeList.getName() + " (" + changeList.getChanges().size() + ")", "", CHANGES_ICON);
      _changeList = changeList;
    }

    public boolean isSelected(AnActionEvent e)
    {
      return _changeList.equals(_defaultChangeList);
    }

    public void setSelected(AnActionEvent e, boolean state)
    {
      if (state) {
        _changeListManager.setDefaultChangeList(_changeList);
      }
    }
  }

  private static void dumpComponentHierarchy(int level, Component component)
  {
    final StringWriter writer = new StringWriter();
    for (int i = 0; i < level - 1; i++) {
      writer.append("| ");
    }
    if (level > 0) {
      writer.append("+-");
    }
    writer.append(component.toString());
    System.out.println(writer.toString());
    if (component instanceof Container) {
      for (final Component child : ((Container)component).getComponents()) {
        dumpComponentHierarchy(level + 1, child);
      }
    }
  }
}