package com.anecdote.ideaplugins.changesbar;

/**
 * Copyright 2006 Nathan Brown
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.intellij.idea.ActionsBundle;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.vcs.VcsBundle;
import com.intellij.openapi.vcs.actions.ShowMessageHistoryAction;
import com.intellij.openapi.vcs.changes.*;
import com.intellij.openapi.vcs.changes.ui.*;
import com.intellij.openapi.vcs.ui.Refreshable;
import com.intellij.openapi.vfs.*;
import com.intellij.openapi.wm.*;
import com.intellij.ui.*;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.*;
import javax.swing.tree.TreeCellRenderer;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.*;
import java.util.List;

public class ChangesBarProjectComponent implements ProjectComponent, VirtualFileListener, ChangeListListener,
  ChangeListener
{
  private final Project _project;
  private ChangesBar _changesBar;
  private StatusBar _statusBar;
  private ChangeListManager _changeListManager;
  private static final Icon COMMIT_ICON = IconLoader.getIcon("/actions/commit.png");
  private static final Icon MESSAGE_HISTORY_ICON = IconLoader.getIcon("/runConfigurations/testOuput.png");
  private static final String CONFLICT_ARROW = ">!!<";
  private static final String DESCRIPTION_START = "Changelist holding the files that were originally" +
                                                  " modified in Changelist \"";
  private static final String DESCRIPTION_MIDDLE = "\" but were subsequently modified in Changelist";
  private static final String DESCRIPTION_END = ".\n\n(Autogenerated by " +
                                                "the ChangesBar plugin.)";
  private ChangeList _defaultChangeList;
  private ChangeList _selectionChangeList;
  static final String CHANGES_TOOLWINDOW_NAME = "Changes";
  private final ShowMessageHistoryAction _messageHistoryAction = new ShowMessageHistoryAction();
  private final JTextArea _commentEditor = new JTextArea();
  private final TitledBorder _commentEditorBorder = new TitledBorder("Comment");
  private boolean _updatingComment;
  private final JButton _commitButton = new JButton(null, COMMIT_ICON);
  private final JButton _messageHistoryButton = new JButton(null, MESSAGE_HISTORY_ICON);
  private ChangesListView _changesListView;
  private Container _changesListViewContainerPanel;
  private JSplitPane _changesListViewCommentViewSplit;
  private Container _changesListViewScrollPane;
  private boolean _commentEditorInstalled;
  private final TreeSelectionListener _changesListViewSelectionListener = new TreeSelectionListener()
  {
    public void valueChanged(TreeSelectionEvent e)
    {
      updateSelectionChangeList();
    }
  };
  private final PropertyChangeListener _changesListViewRendererChangeListener = new PropertyChangeListener()
  {
    public void propertyChange(PropertyChangeEvent evt)
    {
      if (evt.getNewValue() != null) {
        final TreeCellRenderer treeCellRenderer = (TreeCellRenderer)evt.getNewValue();
        changesListViewRendererChanged(treeCellRenderer);
      }
    }
  };
  private boolean _settingChangesListViewRenderer;

  {
    _messageHistoryButton.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        _updatingComment = true;
        try {
          final CheckinProjectPanelStub checkinProjectPanel = new CheckinProjectPanelStub()
          {
            @Override
            public Project getProject()
            {
              return _project;
            }

            @Override
            public void setCommitMessage(String currentDescription)
            {
              _commentEditor.setText(currentDescription);
            }
          };
          _messageHistoryAction.actionPerformed(
            new AnActionEvent(null,
                              new SimpleDataContext(Refreshable.PANEL, checkinProjectPanel),
                              "ChangesBar", _messageHistoryAction.getTemplatePresentation(), null, 0)
          );
          _commentEditor.selectAll();
          _commentEditor.requestFocusInWindow();
        } finally {
          _updatingComment = false;
        }
      }
    });
    _commitButton.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        CommitChangeListDialog.commitChanges(_project, _selectionChangeList.getChanges(), _selectionChangeList,
                                             _changeListManager.getRegisteredExecutors(), true);
      }
    });
  }

  public ChangesBarProjectComponent(Project project)
  {
    _project = project;
  }

  public void initComponent()
  {
    ToolWindowManager.getInstance(_project).invokeLater(new Runnable()
    {
      public void run()
      {
        stateChanged(null);
      }
    });
  }

  public void disposeComponent()
  {
    // TODO: insert component disposal logic here
  }

  public String getComponentName()
  {
    return "ChangeListProjectComponent";
  }

  public void projectOpened()
  {
    // called when project is opened
    final WindowManager windowManager = WindowManager.getInstance();
    _statusBar = windowManager.getStatusBar(_project);
    _changesBar = new ChangesBar(_project, _statusBar);
    _statusBar.addCustomIndicationComponent(_changesBar);
    final ChangeListManager changeListManager = ChangeListManager.getInstance(_project);
    setChangeListManager(changeListManager);
    LocalFileSystem.getInstance().addVirtualFileListener(this);
    ChangesBarApplicationComponent.addChangeListener(this);
    ToolWindowManager.getInstance(_project).invokeLater(new Runnable()
    {
      public void run()
      {
        stateChanged(null);
      }
    });
  }

  private void installCommentEditor()
  {
    final ToolWindow window = ToolWindowManager.getInstance(_project).getToolWindow(CHANGES_TOOLWINDOW_NAME);
    if (window != null) {
      final ChangesListView listView = findChildOfClass(ChangesListView.class, window.getComponent());
      setChangesListView(listView);
    }
  }

  private void deinstallCommentEditor()
  {
    setChangesListView(null);
  }

  private void setChangesListView(ChangesListView listView)
  {
    if (listView == _changesListView) {
      return;
    }
    if (_changesListView != null) {
      _changesListView.removeTreeSelectionListener(_changesListViewSelectionListener);
      _changesListView.removePropertyChangeListener(JTree.CELL_RENDERER_PROPERTY,
                                                    _changesListViewRendererChangeListener);
      if (_commentEditorInstalled) {
        _changesListViewContainerPanel.remove(_changesListViewCommentViewSplit);
        _changesListViewContainerPanel.add(_changesListViewScrollPane, BorderLayout.CENTER);
        _changesListViewContainerPanel.validate();
        _changesListViewCommentViewSplit = null;
        _changesListViewScrollPane = null;
        _changesListViewContainerPanel = null;
        _commentEditorInstalled = false;
      }
    }
    _changesListView = listView;
    if (_changesListView == null) {
      return;
    }
    _changesListView.addPropertyChangeListener(JTree.CELL_RENDERER_PROPERTY,
                                               _changesListViewRendererChangeListener);
    final TreeCellRenderer cellRenderer = _changesListView.getCellRenderer();
    changesListViewRendererChanged(cellRenderer);
    _changesListView.addTreeSelectionListener(_changesListViewSelectionListener);
    _changesListViewScrollPane = _changesListView.getParent().getParent();
    if (_changesListViewScrollPane instanceof JScrollPane) {
      _changesListViewContainerPanel = _changesListViewScrollPane.getParent();
      if (_changesListViewContainerPanel.getLayout() instanceof BorderLayout) {
        _changesListViewContainerPanel.remove(_changesListViewScrollPane);

        final JScrollPane commentScrollPane = new JScrollPane(_commentEditor,
                                                              ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                                                              ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        commentScrollPane.getViewport().setBackground(_commentEditor.getBackground());
        final JPanel commentPanel = new JPanel(new BorderLayout());
        final JPanel borderPanel = new JPanel(new BorderLayout());
        borderPanel.add(commentScrollPane, BorderLayout.CENTER);
        borderPanel.setBorder(_commentEditorBorder);
        commentPanel.add(borderPanel, BorderLayout.CENTER);

        final JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new BorderLayout());
        buttonPanel.add(_commitButton, BorderLayout.EAST);
        buttonPanel.add(_messageHistoryButton, BorderLayout.WEST);
        final Presentation presentation = new Presentation();
        //noinspection UnresolvedPropertyKey
        String text = ActionsBundle.message("action.ChangesView.Commit.text");
        if (text.startsWith("!")) {
          text = "Comm_it Changes";
        }
        presentation.setText(text + "...", true);
        _commitButton.setText(presentation.getText());
        _commitButton.setMnemonic(presentation.getMnemonic());
        _commitButton.putClientProperty("Quaqua.Button.style", "placard");

        presentation.setText("_Message History...", true);
        _messageHistoryButton.setText(presentation.getText());
        _messageHistoryButton.setMnemonic(presentation.getMnemonic());
        borderPanel.add(buttonPanel, BorderLayout.SOUTH);
        _messageHistoryButton.putClientProperty("Quaqua.Button.style", "placard");

        _changesListViewCommentViewSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _changesListViewScrollPane,
                                                          commentPanel);
        _changesListViewContainerPanel.add(_changesListViewCommentViewSplit, BorderLayout.CENTER);
        _changesListViewContainerPanel.validate();
        _changesListViewCommentViewSplit.setDividerLocation(0.5D);
        _changesListViewCommentViewSplit.setResizeWeight(0.5D);

        _commentEditor.setWrapStyleWord(true);
        _commentEditor.setLineWrap(true);
        _commentEditor.addKeyListener(new KeyAdapter()
        {
          @Override
          public void keyTyped(KeyEvent e)
          {
            if (e.getKeyChar() == KeyEvent.VK_ENTER && e.isControlDown()) {
              _commitButton.doClick();
              e.consume();
            }
          }
        });
        _commentEditor.getDocument().addDocumentListener(new DocumentAdapter()
        {
          protected void textChanged(DocumentEvent e)
          {
            _updatingComment = true;
            try {
              if (!(_selectionChangeList instanceof LocalChangeList)) {
                return;
              }
              final LocalChangeList list = (LocalChangeList)_selectionChangeList;
              list.setName(list.getName());
              list.setComment(_commentEditor.getText());
            } finally {
              _updatingComment = false;
            }
          }
        });
        _commentEditorInstalled = true;
        updateCommentEditor(_selectionChangeList);
      } else {
        System.err.println(
          "ChangesBar - View structure not as expected! (Layout) - Aborting Comment Editor Install.");
      }
    } else {
      System.err.println(
        "ChangesBar - View structure not as expected! (JScrollPane) - Aborting Comment Editor Install.");
    }
  }

  private void changesListViewRendererChanged(TreeCellRenderer treeCellRenderer)
  {
    if (_settingChangesListViewRenderer) {
      return;
    }
    _settingChangesListViewRenderer = true;
    try {
      _changesListView.setCellRenderer(new CustomChangeListNodeRenderer(treeCellRenderer));
    } finally {
      _settingChangesListViewRenderer = false;
    }
  }

  private void updateSelectionChangeList()
  {
    final TreePath[] treePaths = _changesListView.getSelectionPaths();
    ChangesBrowserChangeListNode changeListNode = null;
    if (treePaths != null) {
      for (final TreePath treePath : treePaths) {
        final Object node = treePath.getPathComponent(1);
        if (!(node instanceof ChangesBrowserChangeListNode)) {
          continue;
        }
        if (changeListNode == null) {
          changeListNode = (ChangesBrowserChangeListNode)node;
        } else if (!changeListNode.equals(node)) {
          setSelectionChangeList(null);
          return;
        }
      }
    }
    setSelectionChangeList(changeListNode != null ? changeListNode.getUserObject() : null);
  }

  private static <T> T findChildOfClass(Class<T> cls, Component component)
  {
    if (cls.isInstance(component)) {
      return (T)component;
    }
    if (!(component instanceof Container)) {
      return null;
    }
    for (final Component child : ((Container)component).getComponents()) {
      final T childOfClass = findChildOfClass(cls, child);
      if (childOfClass != null) {
        return childOfClass;
      }
    }
    return null;
  }

  private void setChangeListManager(ChangeListManager changeListManager)
  {
    if (_changeListManager == changeListManager) {
      return;
    }
    if (_changeListManager != null) {
      _changeListManager.removeChangeListListener(this);
    }
    _changeListManager = changeListManager;
    if (_changeListManager != null) {
      _changeListManager.addChangeListListener(this);
    }
    _changesBar.setChangeListManager(_changeListManager);
    _defaultChangeList = _changeListManager != null ? _changeListManager.getDefaultChangeList() : null;
  }

  public void projectClosed()
  {
    ChangesBarApplicationComponent.removeChangeListener(this);
    // called when project is being closed
    setChangeListManager(null);
    _statusBar = null;
    _changesBar = null;
    LocalFileSystem.getInstance().removeVirtualFileListener(this);
  }

  private void fileChanged(VirtualFile file, FileChangeType changeType)
  {
    if (!ChangesBarApplicationComponent.isWatchForCollisions() || changeType == FileChangeType.Copied) {
      return;
    }
    final Change change = _changeListManager.getChange(file);
    final LocalChangeList changeList = change != null ? _changeListManager.getChangeList(change) : null;
    final LocalChangeList defaultChangeList = _changeListManager.getDefaultChangeList();
    if (changeList == null || changeList.equals(defaultChangeList)) {
      return;
    }
    final String originalChangeListName = changeList.getName();
    final String newChangeListName = defaultChangeList.getName();
    final List<String> collisionNames = getCollidingChangeListNames(originalChangeListName);
    if (collisionNames != null && collisionNames.contains(newChangeListName)) {
      // already in the set - abort
      return;
    }
    final String collisionChangeListName = originalChangeListName + CONFLICT_ARROW + newChangeListName;
    LocalChangeList list = _changeListManager.findChangeList(collisionChangeListName);
    if (list == null) {
      String descriptionString;
      if (collisionNames != null) {
        collisionNames.add(newChangeListName);
        descriptionString = getCollisionListDescription(collisionNames);
      } else {
        final StringBuilder description = new StringBuilder(DESCRIPTION_START);
        description.append(originalChangeListName);
        description.append(DESCRIPTION_MIDDLE);
        description.append(" \"").append(newChangeListName).append('\"');
        description.append(DESCRIPTION_END);
        descriptionString = description.toString();
      }
      list = _changeListManager.addChangeList(collisionChangeListName,
                                              descriptionString);
      list.setReadOnly(true);
    }
    _changeListManager.moveChangesTo(list, new Change[]{change});
    _changesBar.setWarningVisible(true);
  }

  private static String getCollisionListDescription(List<String> collisionNames)
  {
    final StringBuilder description = new StringBuilder(DESCRIPTION_START);
    description.append(collisionNames.get(0));
    description.append(DESCRIPTION_MIDDLE);
    description.append("s ");
    for (int i = 1; i < collisionNames.size() - 1; i++) {
      final String subName = collisionNames.get(i);
      description.append('\"');
      description.append(subName);
      description.append('\"');
      if (i < collisionNames.size() - 2) {
        description.append(", ");
      }
    }
    description.append(" and \"").append(collisionNames.get(collisionNames.size() - 1)).append('\"');
    description.append(DESCRIPTION_END);
    return description.toString();
  }

  public void contentsChanged(VirtualFileEvent event)
  {
    fileChanged(event.getFile(), FileChangeType.Modified);
  }

  public void fileDeleted(VirtualFileEvent event)
  {
    fileChanged(event.getFile(), FileChangeType.Deleted);
  }

  public void fileMoved(VirtualFileMoveEvent event)
  {
    fileChanged(event.getFile(), FileChangeType.Moved);
  }

  public void fileCopied(VirtualFileCopyEvent event)
  {
    fileChanged(event.getFile(), FileChangeType.Copied);
  }

  public void propertyChanged(VirtualFilePropertyEvent event)
  {
    if (event.getPropertyName().equals(VirtualFile.PROP_NAME)) {
      fileChanged(event.getFile(), FileChangeType.Renamed);
    }
  }

  public void beforePropertyChange(VirtualFilePropertyEvent event)
  {
    // no op
  }

  public void beforeContentsChange(VirtualFileEvent event)
  {
    // no op
  }

  public void beforeFileDeletion(VirtualFileEvent event)
  {
    // no op
  }

  public void beforeFileMovement(VirtualFileMoveEvent event)
  {
    // no op
  }

  public void fileCreated(VirtualFileEvent event)
  {
    // no op
  }

  public void changeListAdded(ChangeList list)
  {
    // no op
  }

  public void changeListRemoved(ChangeList list)
  {
    if (isConflictChangeList(list)) {
      _changesBar.setWarningVisible(false);
    }
  }

  public void changeListRenamed(ChangeList renamedList, String oldName)
  {
    if (isConflictChangeList(renamedList)) {
      return;
    }
    // rename any autogenerated change lists accordingly
    final List<LocalChangeList> localChangeLists = _changeListManager.getChangeLists();
    for (final LocalChangeList changeList : localChangeLists) {
      if (changeList == renamedList) {
        continue;
      }
      final String name = changeList.getName();
      final List<String> list = getCollidingChangeListNames(name);
      if (list == null) {
        continue;
      }
      for (int i = 0; i < list.size(); i++) {
        final String s = list.get(i);
        if (s.equals(oldName)) {
          list.set(i, renamedList.getName());
        }
      }
      changeList.setName(getCollisionListName(list));
      changeList.setComment(getCollisionListDescription(list));
    }
  }

  public void changeListCommentChanged(ChangeList list, String oldComment)
  {
    if (list.equals(_selectionChangeList) && !_updatingComment) {
      updateCommentEditor(list);
    }
  }

  private void updateCommentEditor(ChangeList list)
  {
    _updatingComment = true;
    try {
      _commentEditor.setText(list != null ? list.getComment() : "No list selected");
      _commentEditorBorder.setTitle(list != null ? "Comment for \"" + list.getName() + '\"' : "Comment");
      _commentEditor.setEnabled(list instanceof LocalChangeList);
      _changesListViewContainerPanel.repaint();
      _commitButton.setEnabled(list != null);
    } finally {
      _updatingComment = false;
    }
  }

  private static String getCollisionListName(List<String> collisionNames)
  {
    //noinspection StringBufferWithoutInitialCapacity
    final StringBuilder newName = new StringBuilder();
    for (int i = 0; i < collisionNames.size(); i++) {
      final String subName = collisionNames.get(i);
      newName.append(subName);
      if (i < collisionNames.size() - 1) {
        newName.append(CONFLICT_ARROW);
      }
    }
    return newName.toString();
  }

  public void changesMoved(Collection<Change> changes, ChangeList fromList, ChangeList toList)
  {
    if (changes.isEmpty()) {
      return;
    }
    if (isConflictChangeList(fromList)) {
      _changesBar.setWarningVisible(false);
    }
    removeChangeListIfNotValid(fromList);
  }

  private void removeChangeListIfNotValid(final ChangeList changeList)
  {
    SwingUtilities.invokeLater(new Runnable()
    {
      public void run()
      {
        if (isConflictChangeList(changeList) && changeList.getChanges().isEmpty()) {
          _changeListManager.removeChangeList((LocalChangeList)changeList);
        }
      }
    });
  }

  public void changeListChanged(ChangeList list)
  {
    removeChangeListIfNotValid(list);
  }

  public void defaultListChanged(ChangeList newDefaultList)
  {
    final LocalChangeList oldDefaultList = (LocalChangeList)_defaultChangeList;
    // revert - can't have autogen lists default
    if (isConflictChangeList(newDefaultList)) {
      SwingUtilities.invokeLater(new Runnable()
      {
        public void run()
        {
          _changeListManager.setDefaultChangeList(oldDefaultList);
        }
      }
      );
    } else {
      _defaultChangeList = newDefaultList;
    }
  }

  static boolean isConflictChangeList(ChangeList changeList)
  {
    return changeList.getName().contains(CONFLICT_ARROW);
  }

  public void unchangedFileStatusChanged()
  {
    // to do : work out what this means
//    System.out.println("ChangesBarProjectComponent.unchangedFileStatusChanged()");
  }

  public void changeListUpdateDone()
  {
    // no op
  }

  private static List<String> getCollidingChangeListNames(String changeListCollisionName)
  {
    List<String> result = null;
    if (!changeListCollisionName.contains(CONFLICT_ARROW)) {
      return result;
    }
    final StringTokenizer tokenizer = new StringTokenizer(changeListCollisionName, CONFLICT_ARROW);
    result = new ArrayList<String>();
    while (tokenizer.hasMoreElements()) {
      final String s = (String)tokenizer.nextElement();
      result.add(s);
    }
    return result;
  }

  private void setSelectionChangeList(ChangeList selectionChangeList)
  {
    if (selectionChangeList != null) {
      selectionChangeList = _changeListManager.findChangeList(selectionChangeList.getName());
    }
    _selectionChangeList = selectionChangeList;
    updateCommentEditor(_selectionChangeList);
  }

  public void stateChanged(ChangeEvent e)
  {
    if (_commentEditorInstalled == ChangesBarApplicationComponent.isAddCommentEditor()) {
      return;
    }
    if (ChangesBarApplicationComponent.isAddCommentEditor()) {
      installCommentEditor();
    } else {
      deinstallCommentEditor();
    }
  }

  private static class SimpleDataContext implements DataContext
  {
    private final String _id;
    private final Object _data;

    private SimpleDataContext(String id, Object data)
    {
      _id = id;
      _data = data;
    }

    @Nullable
    public Object getData(@NonNls String dataId)
    {
      if (dataId.equals(_id)) {
        return _data;
      }
      return null;
    }
  }

  private static class CustomChangeListNodeRenderer implements TreeCellRenderer
  {
    private final TreeCellRenderer _cellRenderer;

    private CustomChangeListNodeRenderer(TreeCellRenderer cellRenderer)
    {
      _cellRenderer = cellRenderer;
    }

    public Component getTreeCellRendererComponent(JTree tree, Object value, boolean selected, boolean expanded,
                                                  boolean leaf, int row, boolean hasFocus)
    {
      final Component component = _cellRenderer.getTreeCellRendererComponent(tree, value, selected,
                                                                             expanded, leaf, row,
                                                                             hasFocus);
      if (!(component instanceof SimpleColoredComponent)) {
        return component;
      }
      final SimpleColoredComponent cellRendererComponent =
        (SimpleColoredComponent)component;
      final ChangesBrowserNode node = (ChangesBrowserNode)value;
      if (!(node.getUserObject() instanceof ChangeList)) {
        return component;
      }
      final ChangeList changeList = (ChangeList)node.getUserObject();
      if (!isConflictChangeList(changeList)) {
        return component;
      }
      if (!selected) {
        cellRendererComponent.clear();
        cellRendererComponent.append(changeList.getName(), SimpleTextAttributes.ERROR_ATTRIBUTES);
        appendNodeCount(cellRendererComponent, node);
      }
      cellRendererComponent.setIcon(ChangesBar.WARNING_ICON);
      return component;
    }
  }

  private static void appendNodeCount(SimpleColoredComponent component, ChangesBrowserNode node)
  {
    final int i = node.getCount();
    final int j = node.getDirectoryCount();
    if (j == 0) {
      component.append(new StringBuilder().append(" ").append(VcsBundle.message("changes.nodetitle.changecount",
                                                                                new Object[]{Integer.valueOf(
                                                                                  i)})).toString(),
                       SimpleTextAttributes.GRAY_ITALIC_ATTRIBUTES);
    } else if (i == 0 && j > 0) {
      component.append(new StringBuilder().append(" ").append(VcsBundle.message(
        "changes.nodetitle.directory.changecount", new Object[]{Integer.valueOf(j)})).toString(),
                       SimpleTextAttributes.GRAY_ITALIC_ATTRIBUTES);
    } else {
      component.append(new StringBuilder().append(" ").append(VcsBundle.message(
        "changes.nodetitle.directory.file.changecount", new Object[]{Integer.valueOf(j), Integer.valueOf(
        i)})).toString(), SimpleTextAttributes.GRAY_ITALIC_ATTRIBUTES);
    }
  }
}
