package com.anecdote.ideaplugins.changesbar;

import com.intellij.openapi.util.IconLoader;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.RoundRectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * Copyright 2006 Nathan Brown
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

class ConflictPopup extends JPanel
{
  private static final int BORDER_THICKNESS = 1;
  private static final int SHADOW_DROP = 5;
  private static final float BORDER_TRANS = 0.8F;
  private static final float SHADOW_TRANS = 0.2F;
  private static final int CORNER_RADIUS = 10;
  private Action _clickAction;
  private static final Icon HIDE_ICON = IconLoader.getIcon("/com/anecdote/ideaplugins/changesbar/hide.gif");
  private final JButton _titleButton = new JButton(HIDE_ICON);
  private static final int TRIANGLE_DIMENSION = 18;
  private static final int TRIANGLE_HEIGHT = TRIANGLE_DIMENSION;
  private static final int TRIANGLE_WIDTH = TRIANGLE_DIMENSION;
  private static final int TRIANGLE_INSET = CORNER_RADIUS + 3;
  private static final Insets INSETS = new Insets(5, 5, 5, 5);

  {
    setOpaque(false);
    updateBorder();
  }

  /**
   * Creates a balloon tip with the parameters provided.
   *
   * @param titleText   The text of the title string
   * @param titleIcon   The icon for the title string
   * @param contentText the text contents of the baloon
   */
  private ConflictPopup(String titleText, Icon titleIcon, String contentText)
  {
    super(new GridBagLayout());
    setBackground(UIManager.getColor("info"));
    setForeground(UIManager.getColor("infoText"));

    final JLabel titleLabel = new JLabel(titleText, titleIcon, SwingConstants.LEFT);
    final Font baseFont = titleLabel.getFont();
    titleLabel.setFont(baseFont.deriveFont(Font.BOLD));
    add(titleLabel, new GridBagConstraints(0, 0, 1, 1, 1.0F, 0.0F, GridBagConstraints.WEST,
                                           GridBagConstraints.HORIZONTAL, INSETS, 0, 0));

    _titleButton.setOpaque(false);
    _titleButton.setBorder(BorderFactory.createLineBorder(getForeground()));
    _titleButton.setFocusPainted(false);
    add(_titleButton, new GridBagConstraints(1, 0, 1, 1, 0.0F, 0.0F, GridBagConstraints.CENTER,
                                             GridBagConstraints.NONE, INSETS, 0,
                                             0));

    final JTextArea contentArea = new JTextArea(contentText);
//    _contentArea = new JLabel("<html>" + ComponentUtilities.createFormattedHTML(contentText, null, baseFont, false));
    contentArea.setFont(baseFont);
    contentArea.setBackground(null);
    contentArea.setBorder(null);
    contentArea.setLineWrap(true);
    contentArea.setWrapStyleWord(true);
    contentArea.setEditable(false);
    contentArea.setCursor(Cursor.getDefaultCursor());
    contentArea.setOpaque(false);
    add(contentArea, new GridBagConstraints(0, 1, 2, 1, 1.0F, 1.0F, GridBagConstraints.NORTHWEST,
                                            GridBagConstraints.BOTH, INSETS, 0,
                                            0));

    final MouseAdapter mouseAdapter = new MouseAdapter()
    {
      @Override
      public void mousePressed(MouseEvent e)
      {
        if (_clickAction != null) {
          _clickAction.actionPerformed(new ActionEvent(this, 0, null));
        }
      }
    };
    contentArea.addMouseListener(mouseAdapter);
    titleLabel.addMouseListener(mouseAdapter);
    addMouseListener(mouseAdapter);
  }

  /**
   * Pops up a balloon tip from the specified location on the specified component.
   *
   * @param invoker     the component to pop the balloon up for
   * @param titleText   The text of the title string
   * @param titleIcon   The icon for the title string
   * @param contentText the text contents of the baloon
   * @param clickAction the action that should be executed if the bubble is clicked
   * @param timeout     the time that the bubble should show for before hiding itself
   */
  static InternalPopupPane popup(Component invoker, String titleText, Icon titleIcon,
                                 String contentText, final Action clickAction, int timeout)
  {
    final Point componentPoint = getComponentPoint(invoker);
    final ConflictPopup popup = new ConflictPopup(titleText, titleIcon, contentText);
    final InternalPopupPane popupPane = new InternalPopupPane(popup);
    popup._titleButton.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        popupPane.setVisible(false);
      }
    });
    popup.pack();
    Timer timer = null;
    if (timeout > 0) {
      timer = new Timer(timeout, new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          popupPane.setVisible(false);
        }
      });
      final Timer finalTimer = timer;
      popupPane.addPropertyChangeListener(InternalPopupPane.VISIBLE, new PropertyChangeListener()
      {
        public void propertyChange(PropertyChangeEvent evt)
        {
          if (evt.getNewValue().equals(Boolean.FALSE)) {
            finalTimer.stop();
          }
        }
      });
    }
    popup.setClickAction(new AbstractAction()
    {
      public void actionPerformed(ActionEvent e)
      {
        popupPane.setVisible(false);
        if (clickAction != null) {
          clickAction.actionPerformed(e);
        }
      }
    });
    final Point originPoint = popup.getOriginPoint();
    popupPane.show(invoker, componentPoint.x - originPoint.x, componentPoint.y - originPoint.y);
    popupPane.setSize(popup.getSize());
    popup._titleButton.requestFocus();
    if (timer != null) {
      timer.start();
    }
    return popupPane;
  }

  private void pack()
  {
    setSize(getPreferredSize());
    validate();
    doLayout();
    setSize(getPreferredSize());
    validate();
  }

  private Point getOriginPoint()
  {
    return new Point(getWidth() - (TRIANGLE_INSET + BORDER_THICKNESS + 2 + SHADOW_DROP), getHeight() - (
      BORDER_THICKNESS + 2 + SHADOW_DROP));
  }

  private void setClickAction(Action clickAction)
  {
    _clickAction = clickAction;
  }

  private static Insets getInsetsForBorder()
  {
    final Insets insetsForBorder = new Insets(BORDER_THICKNESS + 2, BORDER_THICKNESS + 2,
                                              BORDER_THICKNESS + 2 + SHADOW_DROP,
                                              BORDER_THICKNESS + 2 + SHADOW_DROP);
    insetsForBorder.bottom += TRIANGLE_HEIGHT;
    return insetsForBorder;
  }

  private Rectangle getShapeBounds()
  {
    final Rectangle shapeBounds = getBounds();
    shapeBounds.height -= TRIANGLE_HEIGHT;
    return shapeBounds;
  }

  private static void fillShape(Graphics2D g2d, Shape shape, Rectangle shapeBounds)
  {
    g2d.fill(shape);
    final ShapePoints triangleFillShape = getTriangleShapePoints(shapeBounds, true);
    g2d.fillPolygon(triangleFillShape.getxPoints(), triangleFillShape.getyPoints(), triangleFillShape.getNumPoints());
  }

  private void drawBorder(Graphics2D g2d, Shape shape, Rectangle shapeBounds, float alpha)
  {
    Composite composite = g2d.getComposite();
    g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, BORDER_TRANS * alpha));
    try {
      g2d.draw(shape);
    } finally {
      g2d.setComposite(composite);
    }
    ShapePoints triangleFillShape = getTriangleShapePoints(shapeBounds, true);
    g2d.setColor(getBackground());
    g2d.fillPolygon(triangleFillShape.getxPoints(), triangleFillShape.getyPoints(), triangleFillShape.getNumPoints());
    triangleFillShape = getTriangleShapePoints(shapeBounds, false);
    g2d.setColor(getForeground());
    composite = g2d.getComposite();
    g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, BORDER_TRANS * alpha));
    try {
      g2d.drawPolyline(triangleFillShape.getxPoints(), triangleFillShape.getyPoints(),
                       triangleFillShape.getNumPoints());
    } finally {
      g2d.setComposite(composite);
    }
  }

  private static class ShapePoints
  {
    private final int[] _xPoints;
    private final int[] _yPoints;
    private final int _numPoints;

    private ShapePoints(int[] xPoints, int[] yPoints, int numPoints)
    {
      _numPoints = numPoints;
      _xPoints = xPoints;
      _yPoints = yPoints;
    }

    private int getNumPoints()
    {
      return _numPoints;
    }

    private int[] getxPoints()
    {
      return _xPoints;
    }

    private int[] getyPoints()
    {
      return _yPoints;
    }
  }

  private static ShapePoints getTriangleShapePoints(Rectangle shapeBounds, boolean fill)
  {
    final int startX = (shapeBounds.x + shapeBounds.width) - TRIANGLE_INSET;
    final int extension = (fill ? 1 : 0);
    final int endX = startX - (TRIANGLE_WIDTH + extension);
    final int startY = (shapeBounds.y + shapeBounds.height) - extension;
    return new ShapePoints(
      new int[]{startX, startX, endX},
      new int[]{startY, startY + TRIANGLE_HEIGHT, startY}, 3);
  }

  private static Point getComponentPoint(Component invoker)
  {
    return new Point(invoker.getWidth() >> 1, invoker.getHeight() >> 1);
  }

  /**
   * {@inheritDoc} overridden to custom paint
   */
  @Override
  public void paint(Graphics g)
  {
    final Graphics2D g2d = (Graphics2D)g;
    final Object renderingHint = g2d.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
    g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    final Composite composite = g2d.getComposite();
    float alpha = 1.0F;
    if (composite instanceof AlphaComposite) {
      alpha = ((AlphaComposite)composite).getAlpha();
    }
    try {
      g2d.setComposite(
        AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
      super.paint(g);
    } finally {
      g2d.setComposite(composite);
      g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, renderingHint);
    }
  }

  /**
   * {@inheritDoc}
   * <p/>
   * Overridden to custom paint
   */
  @SuppressWarnings({"MagicNumber"})
  @Override
  protected void paintComponent(Graphics g)
  {
    final Graphics2D g2d = ((Graphics2D)g);
    final Rectangle bounds = getShapeBounds();
    bounds.height -= BORDER_THICKNESS + SHADOW_DROP;
    bounds.width -= BORDER_THICKNESS + SHADOW_DROP;
    bounds.grow(-1, -1);
    final Shape shape = new RoundRectangle2D.Float(bounds.x, bounds.y, bounds.width, bounds.height, CORNER_RADIUS,
                                                   CORNER_RADIUS);
    final Composite composite = g2d.getComposite();
    final Rectangle shapeBounds = new Rectangle(bounds);
    bounds.translate(SHADOW_DROP - 2, SHADOW_DROP - 2);
    final Rectangle shadowBounds1 = new Rectangle(bounds);
    final Shape shadowShape1 = new RoundRectangle2D.Float(shadowBounds1.x, shadowBounds1.y, shadowBounds1.width,
                                                          shadowBounds1.height, CORNER_RADIUS,
                                                          CORNER_RADIUS);
    bounds.translate(2, 2);
    final Shape shadowShape2 = new RoundRectangle2D.Float(bounds.x, bounds.y, bounds.width, bounds.height,
                                                          CORNER_RADIUS,
                                                          CORNER_RADIUS);
    g.setColor(Color.black);
    float alpha = 1.0F;
    if (composite instanceof AlphaComposite) {
      alpha = ((AlphaComposite)composite).getAlpha();
    }
    g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (SHADOW_TRANS / 2F) * alpha));
    try {
      fillShape(g2d, shadowShape2, bounds);
      g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, SHADOW_TRANS * alpha));
      fillShape(g2d, shadowShape1, shadowBounds1);
    } finally {
      g2d.setComposite(composite);
    }
    g.setColor(getBackground());
    fillShape(g2d, shape, shapeBounds);
    g.setColor(getForeground());
    g2d.setStroke(new BasicStroke(BORDER_THICKNESS));
    drawBorder(g2d, shape, shapeBounds, alpha);
    super.paintComponent(g);
  }


  private void updateBorder()
  {
    setBorder(new EmptyBorder(getInsetsForBorder()));
  }

  static class InternalPopupPane extends JPanel
  {
    private static final int CENTER = 10;

    /**
     * Put the popup at the top of its target, centered horizontally.
     */
    private static final int NORTH = 11;

    /**
     * Put the popup at the top-right corner of its target.
     */
    private static final int NORTHEAST = 12;

    /**
     * Put the popup on the right side of its target, centered vertically.
     */
    private static final int EAST = 13;

    /**
     * Put the popup at the bottom-right corner of its target.
     */
    private static final int SOUTHEAST = 14;

    /**
     * Put the popup at the bottom of its target, centered horizontally.
     */
    private static final int SOUTH = 15;

    /**
     * Put the popup at the bottom-left corner of its target.
     */
    private static final int SOUTHWEST = 16;

    /**
     * Put the popup on the left side of its target, centered vertically.
     */
    private static final int WEST = 17;

    /**
     * Put the popup at the top-left corner of its target.
     */
    private static final int NORTHWEST = 18;

    /**
     * Slide the popup into view vertically
     */
    private static final int VERTICAL_SLIDE = 1;

    /**
     * Slide the popup into view horizontally
     */
    private static final int HORIZONTAL_SLIDE = 2;

    /**
     * Fade the popup into view
     */
    public static final int FADE_IN = 4;

    private Component _invoker;
    private Component _content;
    private int _desiredLocationX;
    private int _desiredLocationY;
    private int _showEffect = NO_EFFECT;
    private int _anchorPoint;
    private JLayeredPane _layeredPane;
    private static final int SHOW_ANIMATION_THREAD_YIELD_TIME = 5;
    private static final int HIDE_ANIMATION_THREAD_YIELD_TIME = 10;
    private static final int DEFAULT_SHOW_ANIMATION_STEPS = 1;
    private static final int DEFAULT_HIDE_ANIMATION_STEPS = 1;
//    private static final int DEFAULT_SHOW_ANIMATION_STEPS = 20;
    //    private static final int DEFAULT_HIDE_ANIMATION_STEPS = 10;
    private final ComponentListener _componentListener = new ComponentListener()
    {
      public void componentHidden(ComponentEvent e)
      {
        setVisible(false);
      }

      public void componentMoved(ComponentEvent e)
      {
        invokerBoundsChanged();
      }

      public void componentResized(ComponentEvent e)
      {
        invokerBoundsChanged();
      }

      public void componentShown(ComponentEvent e)
      {
      }
    };
    private static final int NO_EFFECT = -1;
    public static final String VISIBLE = "visible";
    private float _opacity = 1.0F;


    private InternalPopupPane()
    {
      super();
      setOpaque(false);
      setLayout(null);
    }

    private InternalPopupPane(Component content)
    {
      this();
      setContent(content);
    }

    public Component getContent()
    {
      return _content;
    }

    private void setContent(Component content)
    {
      _content = content;
      removeAll();
      add(content);
      pack();
    }

    private void show(Component invoker, int x, int y)
    {
      _anchorPoint = 0;
      _show(invoker, x, y, null);
    }

    /**
     * Called if there is a change detected in the invoker's position or size.  By default, reshows against the new
     * location of the invoker if visible and an anchor point is specified.
     */
    private void invokerBoundsChanged()
    {
      if (isVisible() && _anchorPoint != 0) {
        show(_invoker, _anchorPoint, null);
      }
    }

    private void _show(Component invoker, int x, int y, Runnable runnable)
    {
      _invoker = invoker;
      setBackground(_invoker.getBackground());
      setForeground(_invoker.getForeground());
      setFont(_invoker.getFont());
      final Point componentLocation = invoker.getLocationOnScreen();
      final JLayeredPane layeredPane = findLayeredPane(_invoker);
      if (layeredPane == null) {
        return;
      }
      final Point lpLocationOnScreen = layeredPane.getLocationOnScreen();
      componentLocation.translate(-lpLocationOnScreen.x, -lpLocationOnScreen.y);
      int posX = componentLocation.x + x;
      final int rightSide = posX + getWidth();
      final int adj = rightSide - layeredPane.getWidth();
      if (adj > 0) {
        posX -= adj;
      }
      final int posY = componentLocation.y + y;
      setLocation(posX, posY);
      setVisible(true, runnable);
    }

    private void show(Component invoker, int anchorPoint, Runnable runnable)
    {
      _anchorPoint = anchorPoint;
      invoker.addComponentListener(_componentListener);
      final Dimension size = getPreferredSize();
      int x = 0;
      int y = 0;
      if (_anchorPoint == NORTH || _anchorPoint == NORTHEAST || _anchorPoint == NORTHWEST) {
        y = -size.height;
        if (_anchorPoint == NORTHEAST) {
          x = invoker.getWidth() - size.width;
        } else if (_anchorPoint == NORTHWEST) {
          x = 0;
        } else { // NORTH
          x = (invoker.getWidth() >> 1) - (size.width >> 1);
        }
      } else if (_anchorPoint == SOUTHEAST || _anchorPoint == SOUTH || _anchorPoint == SOUTHWEST) {
        y = invoker.getHeight();
        if (_anchorPoint == SOUTHEAST) {
          x = invoker.getWidth() - size.width;
        } else if (_anchorPoint == SOUTH) {
          x = (invoker.getWidth() >> 1) - (size.width >> 1);
        } else {
          x = 0;
        }
      }
      if (_anchorPoint == WEST || _anchorPoint == EAST || _anchorPoint == CENTER) {
        y = (invoker.getHeight() - size.height) >> 1;
        if (_anchorPoint == WEST) {
          x = -size.width;
        } else if (_anchorPoint == EAST) {
          x = invoker.getWidth();
        } else { // CENTER
          x = (invoker.getWidth() - size.width) >> 1;
        }
      }
      _show(invoker, x, y, runnable);
    }

    @Override
    public void setLocation(int x, int y)
    {
      final int oldX = _desiredLocationX;
      final int oldY = _desiredLocationY;

      _desiredLocationX = x;
      _desiredLocationY = y;
      if (_layeredPane != null && (x != oldX || y != oldY)) {
        doShow(new Point(x, y), null);
      }
    }

    public int getShowEffect()
    {
      return _showEffect;
    }

    public void setShowEffect(int showEffect)
    {
      _showEffect = showEffect;
    }

    private void pack()
    {
      if (_layeredPane != null) {
        final Dimension pref = getPreferredSize();

        if (pref == null || pref.width != getWidth() ||
            pref.height != getHeight()) {
          show(_layeredPane, _desiredLocationX, _desiredLocationY);
        } else {
          validate();
        }
      } else {
        setSize(getPreferredSize());
      }
    }

    @Override
    public boolean isVisible()
    {
      return _layeredPane != null;
    }

    @Override
    public void setVisible(boolean aFlag)
    {
      setVisible(aFlag, null);
    }

    private void setVisible(boolean aFlag, final Runnable callback)
    {
      // Is it a no-op?
      if (aFlag == isVisible()) {
        if (callback != null) {
          callback.run();
        }
        return;
      }

      if (aFlag) {
        final JLayeredPane layeredPane = findLayeredPane(_invoker);
        if (layeredPane != null) {
          show(layeredPane, _desiredLocationX, _desiredLocationY, callback);
          layeredPane.validate();
          layeredPane.repaint();
        }
        firePropertyChange(VISIBLE, Boolean.FALSE, Boolean.TRUE);
        if (callback != null) {
          callback.run();
        }
      } else if (_layeredPane != null) {
        final PopupFader popupFader = new PopupFader(true, new Runnable()
        {
          public void run()
          {
            try {
              final JLayeredPane layeredPane = _layeredPane;
              if (layeredPane == null) {
                return;
              }
              layeredPane.remove(InternalPopupPane.this);
              layeredPane.validate();
              layeredPane.repaint();
              layeredPane.getParent().invalidate();
              layeredPane.getParent().validate();
              layeredPane.getParent().repaint();
              _layeredPane = null;
            } finally {
              if (callback != null) {
                callback.run();
              }
            }
          }
        });
        _invoker.removeComponentListener(_componentListener);
        firePropertyChange(VISIBLE, Boolean.TRUE, Boolean.FALSE);
        final Thread thread = new Thread(popupFader);
        thread.setPriority(Thread.NORM_PRIORITY - 1);
        thread.start();
//        popupFader.run();
      }
    }

    private void show(JLayeredPane layeredPane, int desiredLocationX, int desiredLocationY, Runnable runnable)
    {
      if (_layeredPane != null && layeredPane != _layeredPane) {
        _layeredPane.remove(this);
      }
      _layeredPane = layeredPane;

      // adjust the location of the popup
//    Point p = adjustPopupLocationToFitInPane(_layeredPane, desiredLocationX, desiredLocationY);
      final Point p = new Point(desiredLocationX, desiredLocationY);
      _desiredLocationX = p.x;
      _desiredLocationY = p.y;

      doShow(p, runnable);
    }

    private void doShow(Point p, Runnable callback)
    {
//    System.out.println("UInternalPopupPane.doShow : p = " + p);
      setOpacity(1.0F);
      final PopupSlider popupSlider = new PopupSlider(p, 1, null);
      setOpacity(0.0F);
      popupSlider.run();
      new Thread(new PopupFader(false, callback)).start();
    }

    /**
     * Returns an point which has been adjusted to take into account of the desktop bounds, taskbar and multi-monitor
     * configuration.
     * <p/>
     * This adustment may be cancelled by invoking the application with -Djavax.swing.adjustPopupLocationToFit=false
     */
    Point adjustPopupLocationToFitInPane(JLayeredPane layeredPane, int xposition, int yposition)
    {
      final Point p = new Point(xposition, yposition);

      final Rectangle panelBounds = layeredPane.getBounds();
      final int scrWidth = panelBounds.width;
      final int scrHeight = panelBounds.height;

      final Dimension size = getPreferredSize();

      if (p.x + size.width > panelBounds.x + scrWidth) {
        p.x = (panelBounds.x + scrWidth) - size.width;
      }

      if (p.y + size.height > panelBounds.y + scrHeight) {
        p.y = (panelBounds.y + scrHeight) - size.height;
      }

      /* Change is made to the desired (X,Y) values, when the
         PopupMenu is too tall OR too wide for the screen
      */
      if (p.x < panelBounds.x) {
        p.x = panelBounds.x;
      }
      if (p.y < panelBounds.y) {
        p.y = panelBounds.y;
      }

      return p;
    }

    @Override
    public Dimension getPreferredSize()
    {
      final Dimension preferredSize = _content.getPreferredSize();
      final Insets insets = getInsets();
      preferredSize.width += insets.left + insets.right;
      preferredSize.height += insets.top + insets.bottom;
      return preferredSize;
    }

    private class PopupFader implements Runnable
    {
      private final boolean _out;
      private final Runnable _callback;

      private PopupFader(boolean out, Runnable callback)
      {
        _out = out;
        _callback = callback;
      }

      public void run()
      {
        try {
          Thread.sleep(_out ? HIDE_ANIMATION_THREAD_YIELD_TIME : SHOW_ANIMATION_THREAD_YIELD_TIME);
        } catch (InterruptedException ex) {
          // no action needed
        }
        final int steps = _out ? DEFAULT_HIDE_ANIMATION_STEPS : DEFAULT_SHOW_ANIMATION_STEPS;
        for (int i = 0; i < steps; i++) {
          final float delta = 1.0F / steps;
          final float amount = (i * delta);
          final float alpha = _out ? 1.0F - amount : 0.0F + amount;
          SwingUtilities.invokeLater(new Runnable()
          {
            public void run()
            {
              setOpacity(alpha);
            }
          });
          try {
            Thread.sleep(HIDE_ANIMATION_THREAD_YIELD_TIME);
          } catch (InterruptedException ex) {
            // no action needed
          }
        }
        SwingUtilities.invokeLater(new Runnable()
        {
          public void run()
          {
            setOpacity(_out ? 0.0F : 1.0F);
            if (_callback != null) {
              _callback.run();
            }
          }
        });
      }
    }

    private class PopupSlider implements Runnable
    {
      private final Point _popupPoint;
      private final int _steps;
      private final Runnable _runnable;

      private PopupSlider(Point p, int steps, Runnable runnable)
      {
        _popupPoint = p;
        _steps = steps;
        _runnable = runnable;
      }

      @SuppressWarnings({"MagicNumber"})
      public void run()
      {
        final Insets insets = getInsets();
        final Dimension targetSize = getPreferredSize();
        final int animationSteps = _steps;
        float dh = 0.0F;
        final int verticalPadding = (insets.top + insets.bottom);
        final float contentHeight = targetSize.height - verticalPadding;
        final int horizontalPadding = (insets.left + insets.right);
        final float contentWidth = targetSize.width - horizontalPadding;
//      if ((_effect & VERTICAL_SLIDE) != 0) {
//        animationSteps = Math.min(targetSize.height, animationSteps);
//      }
//      if ((_effect & HORIZONTAL_SLIDE) != 0) {
//        animationSteps = Math.min(targetSize.width, animationSteps);
//      }
        if ((_showEffect & VERTICAL_SLIDE) != 0) {
          dh = contentHeight / animationSteps;
        }
        float dw = 0.0F;
        if ((_showEffect & HORIZONTAL_SLIDE) != 0) {
          dw = contentWidth / animationSteps;
        }
        for (int i = 0; i < animationSteps + 1; i++) {
          int popupWidth = targetSize.width;
          int popupHeight = targetSize.height;
          if ((_showEffect & VERTICAL_SLIDE) != 0) {
            popupHeight = Math.round((float)i * dh) + verticalPadding;
          }
          if ((_showEffect & HORIZONTAL_SLIDE) != 0) {
            popupWidth = Math.round((float)i * dw) + horizontalPadding;
          }

          int popupX = _popupPoint.x;
          int popupY = _popupPoint.y;

          if (_anchorPoint == NORTHEAST ||
              _anchorPoint == WEST ||
              _anchorPoint == SOUTHEAST) {
            popupX += targetSize.width - popupWidth;
          }
          if (_anchorPoint == NORTHEAST ||
              _anchorPoint == NORTH ||
              _anchorPoint == NORTHWEST) {
            popupY += targetSize.height - popupHeight;
          }
          if (_anchorPoint == CENTER) {
            popupX += (targetSize.width - popupWidth) >> 1;
            popupY += (targetSize.height - popupHeight) >> 1;
          }

          final int popupWidthToSet = popupWidth;
          final int popupHeightToSet = popupHeight;
          final int popupXToSet = popupX;
          final int popupYToSet = popupY;

          if (_layeredPane.getIndexOf(InternalPopupPane.this) == -1) {
            _layeredPane.add(InternalPopupPane.this, JLayeredPane.POPUP_LAYER);
          }
          final Rectangle bounds = new Rectangle(popupXToSet, popupYToSet, popupWidthToSet, popupHeightToSet);
//            System.out.println("UInternalPopupPane$PopupAnimator.run : bounds = " + bounds);
          setBounds(bounds);

          int x = 0;
          int y = 0;
          if (_anchorPoint == SOUTH || _anchorPoint == SOUTHEAST || _anchorPoint == SOUTHWEST) {
            x = popupWidthToSet - targetSize.width;
            y = popupHeightToSet - targetSize.height;
          } else if (_anchorPoint == EAST) {
            x = popupWidthToSet - targetSize.width;
          } else if (_anchorPoint == CENTER) {
            // use Math.round to stop jiggling
            x = Math.round((popupWidthToSet - targetSize.width) / 2F);
            y = Math.round((popupHeightToSet - targetSize.height) / 2F);
          }
          // shift in by margin
          x += insets.left;
          y += insets.top;
          _content.setBounds(x, y, targetSize.width - horizontalPadding,
                             targetSize.height - verticalPadding);
          _layeredPane.revalidate();
//            _layeredPane.repaint(bounds);
          try {
            Thread.sleep(SHOW_ANIMATION_THREAD_YIELD_TIME);
          } catch (InterruptedException ex) {
            // no action needed
          }
        }
        if (_runnable != null) {
          final Runnable runnable = _runnable;
          SwingUtilities.invokeLater(new Runnable()
          {
            public void run()
            {
              runnable.run();
            }
          });
        }
      }
    }

    public float getOpacity()
    {
      return _opacity;
    }

    private void setOpacity(float opacity)
    {
      _opacity = opacity;
      repaint();
    }

    @Override
    public void paint(Graphics g)
    {
      final Graphics2D g2d = (Graphics2D)g;
      final Object renderingHint = g2d.getRenderingHint(RenderingHints.KEY_ANTIALIASING);
      g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      final Composite composite = g2d.getComposite();
      float alpha = 1.0F;
      if (composite instanceof AlphaComposite) {
        alpha = ((AlphaComposite)composite).getAlpha();
      }
      try {
        g2d.setComposite(
          AlphaComposite.getInstance(AlphaComposite.SRC_OVER, _opacity * alpha));
        super.paint(g);
      } finally {
        g2d.setComposite(composite);
        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, renderingHint);
      }
    }

    private static JLayeredPane findLayeredPane(Component component)
    {
      if (component == null) {
        return null;
      }
      final JRootPane rootPane = SwingUtilities.getRootPane(component);
      if (rootPane != null) {
        return rootPane.getLayeredPane();
      }
      return null;
    }
  }
}
